#include "cnaive.h"

bool naive_chain(void *data) {
  storage_t *st = (storage_t*) data;
  clingo_control_t *C = NULL;
  program_t *P = st->P;
  double *a = st->a, *b = st->b, *c = st->c, *d = st->d;

}
